% Text Mining with MATLAB 2nd Edition
% www.http://www.textmininglab.net/

% CHAPTER 9: Geometrical Models

disp('Example 9.1 ------------------------------------------------------');
clear;                                                              %(9.1a)
load datasets_ch9
whos                                                                %(9.1b)
pause;

disp('Example 9.2 ------------------------------------------------------');
% creates a sparse matrix of appropriate dimensions                 %(9.2a)
tdmtx = sparse(length(ranked_vocabulary),length(verses));
% updates the term-document matrix elements                         %(9.2b)
for k = 1:length(verses)
    [void,termindex]=intersect(ranked_vocabulary,verses(k).vocab);
    tdmtx(termindex,k) = 1;
end
pause;
    
disp('Example 9.3 ------------------------------------------------------');
word_index = find(tdmtx(:,400));                                    %(9.3a)
ranked_vocabulary(word_index)                                       %(9.3b)
verses(400).vocab % which are indeed the words in that verse        %(9.3c)
disp(verses(400).text)                                              %(9.3d)
pause;

disp('Example 9.4 ------------------------------------------------------');
doc_index = find(tdmtx(ranked_vocabulary=='spark',:))               %(9.4a)
verses(doc_index).text                                              %(9.4b)
pause;

disp('Example 9.5 ------------------------------------------------------');
full(sum(tdmtx(ranked_vocabulary=='spark',:)))                      %(9.5a)
full(sum(tdmtx(ranked_vocabulary=='river',:)))                      %(9.5b)
full(sum(tdmtx(ranked_vocabulary=='taxes',:)))                      %(9.5c)
pause;

disp('Example 9.6 ------------------------------------------------------');
w1_idx = ranked_vocabulary=='spark';                                %(9.6a)
w2_idx = ranked_vocabulary=='fire';
% 'spark' & 'fire'                                                  %(9.6b)
doc_index = find(tdmtx(w1_idx,:)&tdmtx(w2_idx,:));
disp(verses(doc_index).text)
% 'spark' & not 'fire'                                              %(9.6c)
doc_index = find(tdmtx(w1_idx,:)&not(tdmtx(w2_idx,:)));
disp(verses(doc_index).text)
pause;

disp('Example 9.7 ------------------------------------------------------');
% copies all verses into a string array                             %(9.7a)
for k=1:length(verses), strdocs(k)=verses(k).text; end
% preprocesses text by removing punctuation and lowercasing it      %(9.7b)
strdocs = lower(erasePunctuation(strdocs));
tokdocs = tokenizedDocument(strdocs); % tokenizes the documents     %(9.7c)
bowdocs = bagOfWords(tokdocs) % constructs a bag-of-words model     %(9.7d)
pause;

disp('Example 9.8 ------------------------------------------------------');
comtx = tdmtx*tdmtx';                                                %(9.8)
pause;

disp('Example 9.9 ------------------------------------------------------');
% be careful... an out-of-memory error might occur                   %(9.9)
ovmtx = tdmtx'*tdmtx;
whos comtx ovmtx    
pause;
%  Name           Size                     Bytes  Class     Attributes
%  comtx      12545x12545               35539296  double    sparse    
%  ovmtx      31103x31103            14483390128  double    sparse    
% 14483390128/35539296 = 407.5317
% sum(sum(comtx>0))/prod(size(comtx))*100 = 1.4074
% size(comtx) = 12545 x 12545
% sum(sum(ovmtx>0))/prod(size(ovmtx))*100 = 93.5703
% size(ovmtx) = 31103 x 31103
% (31103 * 31103) / (12545 * 12545) = 6.1470

disp('Example 9.10 -----------------------------------------------------');
sum(sum(ovmtx>0))/numel(ovmtx)*100                                 %(9.10a)
clear ovmtx                                                        %(9.10b)
pause;

disp('Example 9.11 -----------------------------------------------------');
ovmtx = tdmtx(:,1:10000)'*tdmtx(:,1:10000);                         %(9.11)
pause;

disp('Example 9.12 -----------------------------------------------------');
disp(verses(55).text)                                              %(9.12a)
disp(verses(60).text)                                              %(9.12b)
pause

disp('Example 9.13 -----------------------------------------------------');
full(ovmtx(55,60))                                                  %(9.13)
pause

disp('Example 9.14 -----------------------------------------------------');
intersect(verses(55).vocab,verses(60).vocab)                        %(9.14)
pause

disp('Example 9.15 -----------------------------------------------------');
% gets the upper triangular part of ovmtx excluding the diagonal   %(9.15a)
upmtx = triu(ovmtx,1);
% gets the verses with maximum overlap                             %(9.15b)
[v1,v2] = find(upmtx==max(max(upmtx)))
% computes the vocabulary overlap between v1 y v2                  %(9.15c)
full(ovmtx(v1,v2))
disp(verses(v1).text)                                              %(9.15d)
disp(verses(v2).text)                                              %(9.15e)
pause;

disp('Example 9.16 -----------------------------------------------------');
% vocabulary overlap between v1 and v2 relative to v1              %(9.16a)
full(ovmtx(v1,v2))/length(verses(v1).vocab)*100
% vocabulary overlap between v1 and v2 relative to v2              %(8.16b)
full(ovmtx(v1,v2))/length(verses(v2).vocab)*100
pause;

disp('Example 9.17 -----------------------------------------------------');
% find all zero elements in the vocabulary overlap matrix          %(9.17a)
[temp1,temp2] = find(ovmtx==0);
% retain only those elements above the diagonal                    %(9.17b)
v1 = temp1(temp1>temp2);
v2 = temp2(temp1>temp2);
% there iare quite a few verse pairs without vocabulary overlap    %(9.17c)
whos v1 v2
% let us just consider one pair                                    %(9.17d)
disp(verses(v1(20)).text)
disp(verses(v2(20)).text)  
%clear ovmtx upmtx temp1 temp2 v1 v2
pause;
% (2211364*2)/31103^2*100 = 0.4572

disp('Example 9.18 -----------------------------------------------------');
logfreq = log(ranked_frequencies);                                  %(9.18)
pause;

disp('Example 9.19 -----------------------------------------------------');
docfreq = full(sum(tdmtx,2))';                                     %(9.19a)
logdocs = log(size(tdmtx,2)./(docfreq+1));                         %(9.19b)
pause;

disp('Example 9.20 -----------------------------------------------------');
logratio = (logfreq+1).*(logdocs);                                  %(9.20)
pause;

disp('Example 9.21 -----------------------------------------------------');
wsize = 10; % defines a smoothing window of size 10                %(9.21a)
for k=1:length(logratio)-wsize % smoothes the curves               %(9.21b)
    sdocs(k)=mean(logdocs(k:k+wsize));
    sfreq(k)=mean(logfreq(k:k+wsize));
    sratio(k)=mean(logratio(k:k+wsize));
end
pause;

disp('Example 9.22 -----------------------------------------------------');
hf = figure(2); % creates a new figure                             %(9.22a)
set(hf,'Color',[1 1 1],'Name','Word and Document Frequencies');
x = 1:length(logratio)-wsize; % defines a common index             %(9.22b)
% plots the word frequency curve                                   %(9.22c)
subplot(3,1,1); semilogx(x,sfreq,'.'); 
ax = axis; axis([0,12000,ax(3:4)]); ylabel('TF (term-freq)');
% plots the document frequency curve                               %(9.22d)
subplot(3,1,2); semilogx(x,sdocs,'.'); 
ax = axis; axis([0,12000,ax(3:4)]); ylabel('IDF (inv-doc-freq)');
% plots the ratio of both curves                                   %(9.22e)
subplot(3,1,3); semilogx(x,sratio,'.'); 
ax = axis; axis([0,12000,ax(3:4)]); ylabel('Combined Effect');
xlabel('Vocabulary Word Rank')
pause;

disp('Example 9.23 -----------------------------------------------------');
% creates a sparse matrix of appropriate dimensions                %(9.23a)
tfmtx = sparse(length(ranked_vocabulary),length(verses));
% updates term-frequencies for each document                       %(9.23b)
for k = 1:length(verses)
    [void,index] = intersect(ranked_vocabulary,verses(k).vocab);
    tfmtx(index,k) = verses(k).count;
end
pause;

disp('Example 9.24 -----------------------------------------------------');
% compares vector sum(tfmtx,2) with ranked_frequencies              %(9.24)
sum(abs(ranked_frequencies-full(sum(tfmtx,2))'))
pause;

disp('Example 9.25 -----------------------------------------------------');
for k = 1:length(verses)                                            %(9.25)
    ntfmtx(:,k) = tfmtx(:,k)/sum(tfmtx(:,k));
end
pause;

disp('Example 9.26 -----------------------------------------------------');
% computes inverse document frequencies                             %(9.26)
idfvt = log(size(tfmtx,2)./full(sum(tfmtx>0,2)));
pause;

disp('Example 9.27 -----------------------------------------------------');
% creates a sparse matrix of appropriate dimensions                %(9.27a)
tfidfmtx = sparse(length(ranked_vocabulary),length(verses));
for k = 1:length(verses) % computes the TF-IDF matrix              %(9.27b)
    tfidfmtx(:,k) = tfmtx(:,k).*idfvt;
end
pause;

disp('Example 9.28 -----------------------------------------------------');
disp(verses(2971).text)                                            %(9.28)
pause;

disp('Example 9.29 -----------------------------------------------------');
word_rank = find(ranked_vocabulary=='was')                         %(9.29a)
word_tfidf = full(tfidfmtx(word_rank,2971))
word_rank = find(ranked_vocabulary=='bread')                       %(9.29b)
word_tfidf = full(tfidfmtx(word_rank,2971))
word_rank=find(ranked_vocabulary=='hallowed')                      %(9.29c)
word_tfidf = full(tfidfmtx(word_rank,2971))
pause;

disp('Example 9.30 -----------------------------------------------------');
% get the indices of the words in verse 2971                       %(9.30a)
[void,index] = intersect(ranked_vocabulary,verses(2971).vocab);
[void,order] = sort(index); index = index(order);
% prints out rank, freq, tf, idf and tf-idf for each word          %(9.30b)
word_weights = table;
word_weights.Rank = index;
word_weights.Freq = ranked_frequencies(index)';
word_weights.TF = full(tfmtx(index,2971));
word_weights.IDF = full(idfvt(index));
word_weights.('TF-IDF') = full(tfidfmtx(index,2971));
word_weights.Word = ranked_vocabulary(index)';
disp(word_weights)
pause;

disp('Example 9.31 -----------------------------------------------------');
for k = 1:length(verses)                                            %(9.31)
    ntfidfmtx(:,k) = tfidfmtx(:,k)/norm(tfidfmtx(:,k));
end
pause;

disp('Example 9.32 -----------------------------------------------------');
norm(ntfidfmtx(:,2971))                                            %(9.32)
pause;

disp('Example 9.33 - 9.36 ----------------------------------------------');
disp('Examples 9.33-9.36 are not MATLAB scripts')              %(9.33-9.36)
pause;

disp('Example 9.37 -----------------------------------------------------');
x1 = 392; x2 = 2419; x3 = 30063;                                   %(9.37a)
disp(verses(x1).text)                                              %(9.37b)
disp(verses(x2).text)                                              %(9.37c)
disp(verses(x3).text)                                              %(9.37d)
pause;

disp('Example 9.38 -----------------------------------------------------');
x = [x1,x2,x3,x1];                                                 %(9.38a)
for k=1:3
    % gets the pair of vectors to be compared
    va = full(tfidfmtx(:,x(k)));
    vb = full(tfidfmtx(:,x(k+1)));
    % computes the association scores for the given vector pair
    dice(k) = 2*sum((va>0)&(vb>0))/(sum(va>0)+sum(vb>0));
    jaccard(k) = sum((va>0)&(vb>0))/sum((va>0)|(vb>0));
    cosine(k) = (va'*vb)/norm(va)/norm(vb);
end
[dice;jaccard;cosine]                                              %(9.38b)
pause;

disp('Example 9.39 - 9.41 ----------------------------------------------');
disp('Examples 9.39-9.41 are not MATLAB scripts')              %(9.39-9.41)
pause;

disp('Example 9.42 -----------------------------------------------------');
for k=1:3                                                          %(9.42a)
    va = full(ntfidfmtx(:,x(k)));
    vb = full(ntfidfmtx(:,x(k+1)));
    hamming(k) = sum(not((va>0)==(vb>0)));
    euclidean(k) = norm(va-vb);
    cityblock(k) = sum(abs(va-vb));
end
[hamming;euclidean;cityblock]                                      %(9.42b)
pause;

disp('Example 9.43 -----------------------------------------------------');
cosine_scores = ntfidfmtx(:,392)'*ntfidfmtx;                        %(9.43)
pause;

disp('Example 9.44 -----------------------------------------------------');
hf = figure(3); % creates a new figure                             %(9.44a)
set(hf,'Color',[1 1 1]);
set(hf,'Name','Distribution of similarities with respect to verse 392');
% computes a histogram of similarities (20 beams)                  %(9.44b)
[freq,vals] = hist(cosine_scores,20); 
% generates a bar plot considered log frequencies                  %(9.44c)
hb = bar(vals,log(freq)+1);
% adds the figure labels                                           %(9.44d)
xlabel('Cosine Similarity Scores'); 
ylabel('Log-frequencies + 1');       
pause;

disp('Example 9.45 -----------------------------------------------------');
freq(1)                                                             %(9.45)
pause;

